// Project 51 - DS3231 I2C Real-time Clock

#include <avr/io.h>
#include <util/delay.h>

// variabes to store time and date
uint8_t hours, minutes, seconds, dow, dom, mo, years;	

void I2Cenable()
// enable I2C bus
{
    TWBR = 72; // 100 kHz I2C bus
    TWCR |= (1 << TWEN); // enable I2C on PORTC4 and 5
}

void I2Cwait()
// wait until I2C finishes an operation
{
    // wait until bit TWINT in TWCR is set to 1
    while (!(TWCR & (1<<TWINT))); 
}

void I2CstartWait(unsigned char address)
{
    uint8_t status;
    while (1)
    {
        // send START condition
        TWCR = (1<<TWINT) | (1<<TWSTA) | (1<<TWEN);
    
    	// wait until transmission completed
    	I2Cwait();
    
    	// check value of TWSR, and mask out status bits
    	status = TWSR & 0b11111000;
    	if ((status != 0b00001000) && (status != 0b00010000)) continue;
    
    	// send device address
    	TWDR = address;
    	TWCR = (1<<TWINT) | (1<<TWEN);
    
    	// wait until transmission completed
    	I2Cwait();
    
    	// check value of TWSR, and mask out status bits
    	status = TWSR & 0b11111000;
    	if ((status == 0b00100000 )||(status == 0b01011000)) 
    	{    	    
            // secondary device is busy, so send stop condition to terminate write operation
            TWCR = (1<<TWINT) | (1<<TWEN) | (1<<TWSTO);
	        
            // wait until stop condition is executed and I2C bus released
            while(TWCR & (1<<TWSTO));
            continue;
        }
    	break;
    }
}

void I2Cstop()
// stop I2C bus and release GPIO pins
{
    // clear interrupt, enable I2C, generate stop condition
    TWCR |= (1 << TWINT)|(1 << TWEN)|(1 << TWSTO);
}

void I2Cwrite(uint8_t data)
// send 'data' to I2C bus
{
    TWDR = data;
    TWCR |= (1 << TWINT)|(1 << TWEN);
    I2Cwait();
}

uint8_t I2Cread()
// read incoming byte of data from I2C bus
{
    TWCR |= (1 << TWINT)|(1 << TWEN);
    I2Cwait();
    // incoming byte is placed in TWDR register
    return TWDR;	
}

uint8_t I2CreadACK()
// read incoming byte of data from I2C bus and ACK signal
{
    TWCR |= (1 << TWINT)|(1 << TWEN)|(1 << TWEA);
    I2Cwait();
    // incoming byte is placed in TWDR register
    return TWDR;	
}

void writeMAX7219(char hexdata1, char hexdata2)
// sends two bytes in hexadecimal to MAX7219
{
    PORTB &= ~(1 << PORTB0);    // latch pin low
    SPDR = hexdata1;            // send value of hexdata1
    while(!(SPSR & (1<<SPIF))); // wait for SPI transmission to finish
    SPDR = hexdata2;            // send value of hexdata2
    while(!(SPSR & (1<<SPIF))); // wait for SPI transmission to finish
    PORTB |= (1 << PORTB0);     // latch pin high
}

void blankMAX7219()
// blanks all digits
{
    uint8_t i;
    for (i=1; i<9; i++)	    // blank all digits
    {
        writeMAX7219(i,15);
    }
}

void initMAX7219()
// setup MAX7219 for use
{
    PORTB |= (1 << 0); 		// latch pin high (SS)
    SPCR = 0b01010000; 		// set up SPI bus for MAX7219
    writeMAX7219(0x09,0xFF);    // mode decode for digits (table 4 page 7 - 0x09 + 0xFF)
    writeMAX7219(0x0B,0x07);    // set scan limit to 8 digits - 0x09 + 0xFF)
    writeMAX7219(0x0A,0x01);    // set intensity to 8 - 0x0A + 0x08)
    writeMAX7219(0x0C,0x01);    // mode display on (table 4 page 7 - 0x09 + 0xFF)
    blankMAX7219();
} 

void dispMAX7219(uint8_t digit, uint8_t number, uint8_t dp)
// displays "number" in location "digit" with decimal point on/off
// digit: 1~8 for location 1~8
// number: 0~15 for 0~9, - E, H, L, P, blank
// dp: 1 on, 0 off
{
    if (dp==1)      // add decimal point
    { 
        number = number + 128; 
    } 
    writeMAX7219(digit, number);
}

void numberMAX7219(uint32_t value)    
// Displays a number between 0-99999999 on MAX7219-controlled 8 digit display
{
    uint8_t digits[9];
    uint8_t i = 1;

    for (i=1; i<9; i++)
    {
        digits[i]=15;
    }
	
    i = 1;
    while (value > 0) 			 // continue until value > 0
    {
        digits[i] = value % 10;  // determine and store last digit of number
        value = value / 10;      // divide value by 10
        i++;
    }
    for (i=1; i<9; i++)
    {
        dispMAX7219(i, digits[i],0);
    }
}

uint8_t decimalToBcd(uint8_t val)
// Convert integer to BCD
{
    return((val/10*16)+(val%10));
}

uint8_t bcdToDec(uint8_t val)
// Convert BCD to integer
{
    return((val/16*10)+(val%16));
}

void setTimeDS3231(uint8_t hh, uint8_t mm, uint8_t ss, uint8_t dw, uint8_t dd, uint8_t mo, uint8_t yy)
// set the time on DS3231
{
    I2CstartWait(0xD0); // DS3231 write
    I2Cwrite(0x00);		// start with hours register
    I2Cwrite(decimalToBcd(ss));  // seconds
    I2Cwrite(decimalToBcd(mm));  // minutes
    I2Cwrite(decimalToBcd(hh));  // hours
    I2Cwrite(decimalToBcd(dw));  // day of week
    I2Cwrite(decimalToBcd(dd));  // date
    I2Cwrite(decimalToBcd(mo));	 // month
    I2Cwrite(decimalToBcd(yy));	 // year
    I2Cstop();	
}

void readTimeDS3231()
// retrieve time and date from DS3231
{
    I2CstartWait(0xD0); // DS3231 write
    I2Cwrite(0x00);		// seconds register
    I2CstartWait(0xD1); // DS3231 read
    seconds = bcdToDec(I2CreadACK());	
    minutes = bcdToDec(I2CreadACK());	
    hours = bcdToDec(I2CreadACK());	
    dow = bcdToDec(I2CreadACK());	
    dom = bcdToDec(I2CreadACK());
    mo = bcdToDec(I2CreadACK());
    years = bcdToDec(I2CreadACK());
}

void displayTimeMAX7219()
// display time then date on MAX7219 module
{
    blankMAX7219();
    readTimeDS3231();
    // display seconds 
    if (seconds == 0) 
    {	// display '00'
        dispMAX7219(1,0,0);
        dispMAX7219(2,0,0);
    } else if (seconds >0 && seconds <10)
    {	// display leading zero
        dispMAX7219(1,seconds,0);
        dispMAX7219(2,0,0);	
    } else
    {	// seconds > 10
        numberMAX7219(seconds);
    }

    dispMAX7219(3,10,0); // display a dash
	
    // display minutes
    if (minutes == 0) 
    {	// display '00'
        dispMAX7219(4,0,0);
        dispMAX7219(5,0,0);
    } else if (minutes >0 && minutes <10)
    {	// display leading zero
        dispMAX7219(4,minutes,0);
        dispMAX7219(5,0,0);	
    } else
    {	// minutes > 10
        dispMAX7219(4,(minutes % 10),0);
        dispMAX7219(5,(minutes / 10),0);		
    }	
	
    dispMAX7219(6,10,0); // display a dash	
	
    // display hours
    if (hours == 0) 
    {	// display '00'
        dispMAX7219(7,0,0);
        dispMAX7219(8,0,0);
    } else if (hours >0 && hours <10)
    {	// display leading zero
        dispMAX7219(7,hours,0);
        dispMAX7219(8,0,0);	
    } else
    {	// hours > 10
        dispMAX7219(7,(hours % 10),0);
        dispMAX7219(8,(hours / 10),0);		
    }	

    _delay_ms(1000);

    // display date 
    if (dom >0 && dom <10)
    {	// display leading zero
        dispMAX7219(7,dom,0);
        dispMAX7219(8,0,0);	
	} else
	{	// seconds > 10
            dispMAX7219(8,(dom / 10), 0);
            dispMAX7219(7,(dom % 10), 0);
	}
	
	dispMAX7219(6,10,0); // display a dash
		
	// display month 
	if (mo >0 && mo <10)
	{	// display leading zero
            dispMAX7219(4,mo,0);
            dispMAX7219(5,0,0);	
	} else
	{	// seconds > 10
            dispMAX7219(5,(mo / 10), 0);
            dispMAX7219(4,(mo % 10), 0);
	}
	
	dispMAX7219(3,10,0); // display a dash
	
	// display year
	if (years == 0) 
	{	// display '00'
            dispMAX7219(1,0,0);
            dispMAX7219(2,0,0);
	} else 			
	if (years >0 && years <10)
	{	// display leading zero
            dispMAX7219(1,years,0);
            dispMAX7219(2,0,0);	
	} else
	{	// years > 10
            dispMAX7219(2,(years / 10), 0);
            dispMAX7219(1,(years % 10), 0);
        }	
    _delay_ms(1000);	
}

int main(void)
{
    DDRB = 0b11111111; // set PORTB as outputs	
    I2Cenable();
    initMAX7219();
    // uncomment to set time and date, then comment and re-flash code
    // setTimeDS3231(9,13,0,5,29,4,21); // h,m,s,dow,dom,m,y
    while (1) 
    {
        displayTimeMAX7219();
        _delay_ms(250);
    }
}